#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2001-2004 Free Software Foundation
#
# $Id: GCParser.py 5921 2004-06-21 18:14:44Z johannes $

import re

from gnue.common.definitions import GParser, GObjects, GRootObj
from gnue.common.definitions.GParserHelpers import GContent
from gnue.common.formatting import GTypecast
from gnue.appserver.classrep import helpers, Namespace


xmlElements   = None
_LENGTH_SCALE = re.compile ('^\w+\s*\(\s*(\d*)\s*[\,]{0,1}\s*(\d*)\s*\)\s*')


# =============================================================================
# Exceptions
# =============================================================================

class Error (gException):
  pass

class NoModuleError (Error):
  def __init__ (self, classname):
    msg = u_("Class '%s' has no module") % classname
    Error.__init__ (self, msg)

class DuplicateDefinitionError (Error):
  def __init__ (self, name, asLength):
    if asLength:
      msg = u_("'%s' has a length in type and an extra length attribute") % name
    else:
      msg = u_("'%s' has a scale in type and an extra scale attribute") % name
    Error.__init__ (self, msg)

class ModuleMismatchError (Error):
  def __init__ (self, classname, module):
    msg = u_("The class '%(class)s' mismatches the given module attribute "
             "'%(module)s'") % {'class': classname, 'module': module}
    Error.__init__ (self, msg)

class InvalidCalculatedError (Error):
  def __init__ (self, classname, propName):
    msg = u_("Calculated property '%(class)s.%(property)s' has parameters") \
          % {'class': classname, 'property': propName}
    Error.__init__ (self, msg)

# =============================================================================
# load an XML object tree from a given stream and return it's root object
# =============================================================================

def loadFile (stream, initialize = True):
  return GParser.loadXMLObject (stream, xmlSchemaHandler, 'GCModule',
      'module', initialize)


# =============================================================================
# Build a dictionary tree with all available XML elements
# =============================================================================

def getXMLelements ():
  global xmlElements

  if xmlElements is None:
    xmlElements = {
      'module': {
        'BaseClass' : GCModule,
        'Required'  : True,
        'Attributes': {
          'name': {
            'Required': True,
            'Unique'  : True,
            'Typecast': GTypecast.name},
          'comment': {
            'Typecast': GTypecast.text}
        },
        'ParentTags': (None)
      },

      'class': {
        'BaseClass': GCClass,
        'Attributes': {
          'name': {
            'Required': True,
            'Unique'  : True,
            'Typecast': GTypecast.name},
          'module': {
            'Typecast': GTypecast.name},
          'comment': {
            'Typecast': GTypecast.text}
        },
        'ParentTags': ('module',)
      },

      'property': {
        'BaseClass': GCProperty,
        'Attributes': {
          'name': {
            'Required': True,
            'Unique'  : True,
            'Typecast': GTypecast.name},
          'type': {
            'Required': True,
            'Typecast': GTypecast.text},
          'nullable': {
            'Default' : True,
            'Typecast': GTypecast.boolean},
          'comment': {
            'Typecast': GTypecast.text},
          'length': {
            'Typecast': GTypecast.integer},
          'scale': {
            'Typecast': GTypecast.integer},
          'language': {
            'Default' : 'python',
            'Typecast': GTypecast.name}
        },
        'MixedContent'  : True,
        'KeepWhitespace': True,
        'ParentTags'    : ('class',)
      },

      'procedure': {
        'BaseClass'     : GCProcedure,
        'Attributes'    : {
          'name': {
            'Required': True,
            'Unique'  : True,
            'Typecast': GTypecast.name},
          'type': {
            'Typecast': GTypecast.text},
          'nullable': {
            'Default' : True,
            'Typecast': GTypecast.boolean},
          'comment': {
            'Typecast': GTypecast.text},
          'length': {
            'Typecast': GTypecast.integer},
          'scale': {
            'Typecast': GTypecast.integer},
          'language': {
            'Default' : 'python',
            'Typecast': GTypecast.name}
        },
        'MixedContent'  : True,
        'KeepWhitespace': True,
        'ParentTags'    : ('class',)
      },

      'parameter': {
        'BaseClass': GCParameter,
        'Attributes': {
          'name': {
            'Required': True,
            'Unique'  : True,
            'Typecast': GTypecast.name},
          'type': {
            'Required': True,
            'Typecast': GTypecast.text},
          'comment': {
            'Typecast': GTypecast.text},
          'length': {
            'Typecast': GTypecast.integer},
          'scale': {
            'Typecast': GTypecast.integer}
        },
        'ParentTags': ('procedure',)
      }
    }

  return GParser.buildImportableTags ('module', xmlElements)


# =============================================================================
# Class called by the XML parser to process the XML file
# =============================================================================

class xmlSchemaHandler (GParser.xmlHandler):
  def __init__ (self):
    GParser.xmlHandler.__init__ (self)
    self.xmlElements = getXMLelements ()


# =============================================================================
# Base class for all GNUe Class Definition classes
# =============================================================================

class GCObject (GObjects.GObj):
  pass


# =============================================================================
# Base class for all objects with a type-attribute
# =============================================================================

class GCTypeDefinition (GCObject):
  def __init__ (self, parent = None, objType = None):
    GCObject.__init__ (self, parent, type = objType)

    self.datatype    = None
    self.length      = None
    self.scale       = None
    self.isReference = False

    self._inits.extend ([None, self._validate])


  # ---------------------------------------------------------------------------
  # If an instance has a type attribute, verify it
  # ---------------------------------------------------------------------------

  def _validate (self):

    if hasattr (self, 'type'):
      fieldType = self.type.lower ().strip ()
      tpMatch   = re.compile ('^(\w+)').match (fieldType)
      if tpMatch is None:
        raise helpers.TypeNameError, (fieldType)

      typename = tpMatch.groups () [0]
      if not typename in helpers.BASE_TYPES and len (typename.split ('_')) != 2:
        raise helpers.TypeNameError, (fieldType)

      # try to extract length and scale from fieldType
      lsMatch = _LENGTH_SCALE.match (fieldType)
      if lsMatch is not None:
        (lstr, sstr) = lsMatch.groups ()

        if len (lstr):
          if self.length is not None and self.length:
            raise DuplicateDefinitionError, (self.name, True)
          self.length = int (lstr)

        if len (sstr):
          if self.scale is not None and self.scale:
            raise DuplicateDefinitionError, (self.name, False)
          self.scale = int (sstr)

      if typename in helpers.BASE_TYPES:
        helpers.verifyBasetype (typename, self.length, self.scale)
      else:
        self.isReference = True

        if self.length:
          raise helpers.TypeFormatError, \
              u_("Reference type '%s' must not have a length") % typename
        if self.scale:
          raise helpers.TypeFormatError, \
              u_("Reference type '%s' must not have a scale") % typename

        typename = 'string'
        self.length = 32

      self.datatype = typename


# =============================================================================
# The module object, root object of a GCD tree
# =============================================================================

class GCModule (GRootObj.GRootObj, GCObject):
  def __init__ (self, parent = None):
    GRootObj.GRootObj.__init__ (self, 'module', getXMLelements, self.__module__)
    GCObject.__init__ (self, parent, type = 'GCModule')


# =============================================================================
# The class object
# =============================================================================

class GCClass (GCObject):
  def __init__ (self, parent):
    GCObject.__init__ (self, parent, type = 'GCClass')
    self.fullName = None
    self.action   = 'create'
    self._inits.extend ([None, self._validate, self._complete])

  # ---------------------------------------------------------------------------
  # Validate a class definition
  # ---------------------------------------------------------------------------

  def _validate (self):
    if not isinstance (self._parent, GCModule):
      raise NoModuleError, self.name

    nameParts = self.name.split ('_')
    if len (nameParts) > 1:
      if hasattr (self, 'module') and self.module != nameParts [0]:
        raise ModuleMismatchError, (self.name, self.module)

      (self.module, self.name) = nameParts [:2]

    if not hasattr (self, 'module'):
      self.module = self._parent.name

    # if this class is not defined by the top level module, this is just a
    # definition for a class extension
    if self.module != self._parent.name:
      self.action = 'extend'

    self.fullName = Namespace.createName (self.module, self.name)


  # ---------------------------------------------------------------------------
  # Complete a class definition if neccessary
  # ---------------------------------------------------------------------------

  def _complete (self):
    # check if there's a gnue_id in new classes
    if self.action == 'create':
      items = [c.fullName for c in self.findChildrenOfType ('GCProperty')]
      if not 'gnue_id' in items:
        gnueId = GCProperty (self)
        gnueId.name     = 'id'
        gnueId.type     = 'string'
        gnueId.length   = 32
        gnueId.comment  = 'Object ID'
        gnueId.nullable = False

        gnueId._validate ()
        gnueId.module  = 'gnue'
        gnueId.fullName= 'gnue_id'

    # transform all 'calculated properties' into procedures
    for prop in self.findChildrenOfType ('GCProperty') [:]:
      if True in [isinstance (c, GContent) for c in prop._children]:
        if prop.findChildOfType ('GCParameter'):
          raise InvalidCalculatedError, (self.fullName, prop.fullName)

        proc = GCProcedure (self)
        proc.name      = 'get%s' % prop.name
        proc.type      = prop.datatype
        proc.length    = prop.length
        proc.scale     = prop.scale
        proc.nullable  = prop.nullable
        proc.language  = prop.language
        if hasattr (prop, 'comment'):
          proc.comment = prop.comment

        proc._children = prop._children

        for child in proc._children:
          child._parent = proc

        proc._validate ()

        self._children.remove (prop)

      



# =============================================================================
# The property object
# =============================================================================

class GCProperty (GCTypeDefinition):
  def __init__ (self, parent):
    GCTypeDefinition.__init__ (self, parent, objType = 'GCProperty')
    self.module   = None
    self.fullName = None


  # ---------------------------------------------------------------------------
  # Validate a property definition
  # ---------------------------------------------------------------------------

  def _validate (self):
    GCTypeDefinition._validate (self)
    self.module   = self.findParentOfType ('GCModule').name
    self.fullName = Namespace.createName (self.module, self.name)


# =============================================================================
# The procedure object
# =============================================================================

class GCProcedure (GCTypeDefinition):
  def __init__ (self, parent):
    GCTypeDefinition.__init__ (self, parent, objType = 'GCProcedure')
    self.module   = None
    self.fullName = None


  # ---------------------------------------------------------------------------
  # Validate a procedure definition
  # ---------------------------------------------------------------------------

  def _validate (self):
    GCTypeDefinition._validate (self)
    self.module   = self.findParentOfType ('GCModule').name
    self.fullName = Namespace.createName (self.module, self.name)


# =============================================================================
# The parameter object
# =============================================================================

class GCParameter (GCTypeDefinition):
  def __init__ (self, parent):
    GCTypeDefinition.__init__ (self, parent, objType = 'GCParameter')

