# GNU Enterprise Application Server - List Object
#
# Copyright 2001-2004 Free Software Foundation
#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# $Id: geasList.py 5808 2004-05-12 17:09:43Z johannes $

import geasInstance
import string

# =============================================================================
# List class
# =============================================================================

class geasList:

  # ---------------------------------------------------------------------------
  # Initalize
  # ---------------------------------------------------------------------------

  def __init__ (self, session, classdef, connection, recordset, propertylist,
                condition, dsSort, asSort):
    self.__session    = session
    self.__classdef   = classdef
    self.__connection = connection
    self.__recordset  = recordset
    self.__prefetch   = propertylist      # property names to be prefetched
    self.__condition  = condition
    self.__instances  = []
    self.__unsorted   = []
    self.__isComplete = False
    self.__dsSorting  = dsSort          # sorted by datasource 
    self.__asSorting  = asSort          # additional sortorder of this list

    if len (self.__asSorting):
      self.__fillup = self.__fillupSorted
    else:
      self.__fillup = self.__fillupUnsorted


  # ---------------------------------------------------------------------------
  # fill up the instance list with up to count instances using sorting/grouping
  # ---------------------------------------------------------------------------

  def __fillupSorted (self, count):
    """
    This function fills the internal list of instances with up to count
    elements. If @count is 0 all available instances are added to the list.
    """

    while self.__wantMore (count):

      # if a pending element is available, start with it; otherwise fetch the
      # next one from the datasource
      if len (self.__unsorted):
        current = self.__unsorted.pop ()
      else:
        current = self.__getInstance (len (self.__instances) == 0)

      while self.__wantMore (count) and current is not None:
        group = self.__getGroup (current)

        while current is not None and self.__getGroup (current) == group:
          self.__unsorted.append (current)
          current = self.__getInstance (False)

        if len (self.__unsorted) > 1:
          self.__sortBatch ()

        self.__instances.extend (self.__unsorted)

        # if no more instances are needed at the moment, but there was a valid
        # element at the end, we hold it as a 'pending but still unsorted' one
        if not self.__wantMore (count) and current is not None:
          self.__unsorted = [current]
        else:
          self.__unsorted = []


      if current is None:
        break


  # ---------------------------------------------------------------------------
  # Fill up instance list without any sorting/grouping
  # ---------------------------------------------------------------------------

  def __fillupUnsorted (self, count):
    """
    This function fills up the internal list of instances with up to @count
    elements.
    """

    while self.__wantMore (count):
      current = self.__getInstance (len (self.__instances) == 0)
      if current is not None:
        self.__instances.append (current)


  # ---------------------------------------------------------------------------
  # wanna have more instances ?
  # ---------------------------------------------------------------------------

  def __wantMore (self, count):
    """
    This function returns TRUE if more instances are needed to fill up the
    instance cache with @count elements. If @count is zero, an unlimited amount
    of instances is accepted.
    """
    return not self.__isComplete and \
          (not count or len (self.__instances) < count)



  # ---------------------------------------------------------------------------
  # Get the next usable instance and add it to the instance list
  # ---------------------------------------------------------------------------

  def __getInstance (self, fromStart = False):
    """
    This function fetches the next usable instance from the recordset according
    to the list's condition. This instance will be returned as function result.
    """

    if fromStart:
      record = self.__recordset.firstRecord ()
    else:
      record = self.__recordset.nextRecord ()

    self.__isComplete = record is None

    if record is not None:
      instance = geasInstance.geasInstance (self.__session, self.__connection,
                                            record, self.__classdef)
      if self.__condition is not None:
        if not self.__condition.evaluate (instance):
          return self.__getInstance ()

      return instance

    else:
      return None


  # ---------------------------------------------------------------------------
  # Get a grouping sequence for an instance
  # ---------------------------------------------------------------------------

  def __getGroup (self, instance):
    """
    This function returns a grouping list according to the datasource-sorting.
    If no such sort order is defined, None is returned.
    """
    if len (self.__dsSorting):
      return instance.get (self.__dsSorting)
    else:
      return None


  # ---------------------------------------------------------------------------
  # sort the current batch in __unsorted
  # ---------------------------------------------------------------------------

  def __sortBatch (self):
    """
    This function sorts the unsorted batch according to asSorting.
    """
    if len (self.__asSorting):
      slist = [i.get (self.__asSorting) + [i] for i in self.__unsorted]
      slist.sort ()
      self.__unsorted = [item [-1] for item in slist]


  # ---------------------------------------------------------------------------
  # Get the length of the list (the number of entries)
  # ---------------------------------------------------------------------------

  def count (self):
    """
    This function returns the number of instances available in this list.
    """
    if self.__condition is not None:
      self.__fillup (0)
      return len (self.__instances)
    else:
      return self.__recordset.count ()


  # ---------------------------------------------------------------------------
  # Fetch data from the database backend
  # ---------------------------------------------------------------------------

  def fetch (self, start, count):
    """
    This function returns @count instances skipping the first @start instances.
    If @start is negative, the starting point is from the end of the list.
    """

    if start < 0:
      self.__fillup (0)
    else:
      self.__fillup (start + count)

    result = []
    for pos in range (start, min (start + count, len (self.__instances))):
      result.append (self.__instances [pos].get (self.__prefetch))

    return result


