# GNU Enterprise Application Server - Session Object
#
# Copyright 2001-2004 Free Software Foundation
#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# $Id: geasSession.py 6109 2004-07-30 06:31:35Z johannes $

from types import *

import string
import whrandom

from gnue import appserver
from gnue.appserver import data
from gnue.appserver.geasAuthentication import AuthError

import geasList, geasInstance
from gnue.common.datasources import GConditions

from mx import DateTime

class InstanceNotFoundError (appserver.Error):
  def __init__ (self, classname, objectId):
    msg = u_("Instance '%(objectId)s' of class '%(classname)s' not found") \
          % {'objectId' : objectId,
             'classname': classname }
    appserver.Error.__init__ (self, msg)

class AccessDeniedError (AuthError):
  def __init__ (self, classname):
    msg = u_("Access to class '%s' denied") % classname
    AuthError.__init__ (self, msg)

class ListNotFoundError (appserver.Error):
  def __init__ (self, listId):
    msg = u_ ("Cannot find a list with ID '%s'") % listId
    appserver.Error.__init__ (self, msg)


# =============================================================================
# Session class
# =============================================================================

class geasSession:

  # ---------------------------------------------------------------------------
  # Initalize
  # ---------------------------------------------------------------------------

  def __init__ (self, connections, authAdapter, sm, id):

    self.loggedIn = 0
    self.connections = connections
    try:
      self.database = gConfig('database')
    except:
      self.database = "gnue"

    self.__user           = ""
    self.__lists          = {}
    self.__listcount      = 0
    self.__authAdapter    = authAdapter
    self.__connection     = None
    self.__dirtyInstances = {}
    self.__operation      = None

    self.sm = sm                        # The session manager
    self.id = id                        # The session id

  # ---------------------------------------------------------------------------
  # Get a class definition from a class name and check access
  # ---------------------------------------------------------------------------

  def __getClassdef (self, classname):

    # check if user has access rights for this class
    if not self.__authAdapter.hasAccess (self, self.__user, classname):
      raise AccessDeniedError, (classname)

    return self.sm.classes [classname]



  def __getProcdef (self, classdef, procedurename):
    # add access control to procedures here
    return classdef.procedures [procedurename]

  # ---------------------------------------------------------------------------
  # Get a field name from a property name, resolving references
  # ---------------------------------------------------------------------------

  def __getFieldname (self, classdef, propertyname, contentdict, add,
                      skipCalculated = True):

    elements = string.split (propertyname, '.')

    a = u't0'                           # start with main table
    c = classdef
    for e in elements [:-1]:
      p = c.properties [e]
      if not p.isReference:
        raise geasInstance.ReferenceError, (propertyname, p.fullName)
      c = p.referencedClass
      # do we already have a link?
      found = False
      for (alias, (table, fk_alias, fk_field, fields)) in contentdict.items ():
        if table == c.table and fk_alias == a and fk_field == p.column:
          found = True
          a = alias
          break
      # have to create new alias (new link)
      if not found:
        new_alias = u't' + str (len (contentdict))
        contentdict [new_alias] = (c.table, a, p.column, [])
        a = new_alias

    # the real (final) propertydef
    p = c.properties [elements [-1]]
    if skipCalculated and p.isCalculated:
      return None

    # add new field to fieldlist
    if add:
      (contentdict [a] [3]).append (p.column)

    # return the column name
    return a + '.' + p.column

  # ---------------------------------------------------------------------------
  # Convert property names in conditions to field names
  # ---------------------------------------------------------------------------

  def __convertCondition (self, classdef, condition, contentdict):

    if condition is not None:
      if isinstance (condition, GConditions.GCField):
        condition.name = self.__getFieldname (classdef, condition.name,
                                            contentdict, False)
      else:
        for child in condition._children:
          self.__convertCondition (classdef, child, contentdict)

  # ---------------------------------------------------------------------------
  # Start a new 'atomic' operation if not already started
  # ---------------------------------------------------------------------------

  def __startOperation (self, operation):
    if self.__operation is None:
      self.__operation = operation
      return True
    else:
      return False



  # ---------------------------------------------------------------------------
  # Cancel all changes of an operation
  # ---------------------------------------------------------------------------

  def __cancelOperation (self):
    try:
      self.__connection.cancelChanges ()

    finally:
      self.__operation = None

  # ---------------------------------------------------------------------------
  # Confirm all changes of an operation
  # ---------------------------------------------------------------------------

  def __confirmOperation (self):
    try:
      self.__connection.confirmChanges ()

    finally:
      self.__operation = None


  # ---------------------------------------------------------------------------
  # Log into the application server
  # ---------------------------------------------------------------------------

  def login (self, user, password):

    # This username/password is for the Application Server, not for the
    # database.
    self.__user = user
    self.loggedIn = self.__authAdapter.authenticate(self, user,
                                                   {'password':password} )
    if self.loggedIn:
      self.__connection = data.connection (self.connections, self.database)

    return self.loggedIn

  # ---------------------------------------------------------------------------
  # Logout from the application server
  # ---------------------------------------------------------------------------

  def logout (self):

    # FIXME: should the authAdapter be contacted?
    self.__connection.close ()
    self.loggedIn = 0

  # ---------------------------------------------------------------------------
  # Commit the active transaction
  # ---------------------------------------------------------------------------

  def commit (self):
    for instance in self.__dirtyInstances.values ():
      instance.validate ()

    self.__connection.commit ()
    self.__dirtyInstances = {}

  # ---------------------------------------------------------------------------
  # Rollback the active transaction
  # ---------------------------------------------------------------------------

  def rollback (self):

    self.__connection.rollback ()
    self.__dirtyInstances = {}

  # ---------------------------------------------------------------------------
  # Create a new list of business objects of a given class
  # ---------------------------------------------------------------------------

  def request (self, classname, conditions, sortorder, propertylist):

    classdef = self.__getClassdef (classname)

    (dsCond, asCond) = self.__splitCondition (conditions, classdef)

    # Start to build the content dictionary for self.__connection.query()
    content = {}
    content [u't0'] = (classdef.table, None, None, [u'gnue_id'])

    for p in propertylist:
      self.__getFieldname (classdef, p, content, True)

    if asCond is not None:
      self.__addConditionToContent (classdef, content, asCond)

    if dsCond is not None:
      self.__convertCondition (classdef, dsCond, content)

    sortlist = []     # translated sortlist for datasource
    dsSort   = []     # datasource sortlist
    asSort   = []     # appserver sortlist
    for ix in range (0, len (sortorder)):
      propName  = sortorder [ix]
      fieldName = self.__getFieldname (classdef, propName, content, False)

      if fieldName is None:
        asSort = sortorder [ix:]
        for item in asSort:
          self.__getFieldname (classdef, item, content, True)
        break
      else:
        sortlist.append (fieldName)
        dsSort.append (propName)

    recordset = self.__connection.query (content, dsCond, sortlist)

    list = geasList.geasList (self, classdef, self.__connection, recordset,
                           [u'gnue_id'] + propertylist, asCond, dsSort, asSort)

    self.__listcount += 1
    self.__lists [self.__listcount] = list
    list_id = self.__listcount
    return list_id;

  # ---------------------------------------------------------------------------
  # Check list id and raise exception if invalid, return list otherwise
  # ---------------------------------------------------------------------------

  def __getList (self, list_id):

    if self.__lists.has_key (list_id):
      return self.__lists [list_id]
    else:
      raise ListNotFoundError, (list_id)

  # ---------------------------------------------------------------------------
  # Count the number of objects in the list
  # ---------------------------------------------------------------------------

  def count (self, list_id):

    list = self.__getList (list_id)
    return list.count ();

  # ---------------------------------------------------------------------------
  # Fetch data from the database backend
  # ---------------------------------------------------------------------------

  def fetch (self, list_id, start, count):

    list = self.__getList (list_id)
    return list.fetch (start, count)

  # ---------------------------------------------------------------------------
  # Create a single geasInstance object from an existing record
  # ---------------------------------------------------------------------------

  def __findInstance (self, classdef, object_id, propertylist):

    table = classdef.table
    # Only query the main table here.  geasInstance.get will do the rest.
    fieldlist = []
    for p in propertylist:
      pdef = classdef.properties [string.split (p, '.') [0]]
      if not pdef.isCalculated and not pdef.column in fieldlist:
        fieldlist.append (pdef.column)

    record = self.__connection.findRecord (table, object_id, fieldlist)
    if record is None:
      raise InstanceNotFoundError, (classdef.fullName, object_id)

    return geasInstance.geasInstance (self, self.__connection, record,
                                      classdef)

  # ---------------------------------------------------------------------------
  # Create a single geasInstance object for a new record
  # ---------------------------------------------------------------------------

  def __newInstance (self, classdef):

    table    = classdef.table
    record   = self.__connection.insertRecord (table)
    instance = geasInstance.geasInstance (self, self.__connection, record,
                                          classdef)

    for proc in classdef.procedures.values ():
      if proc.gnue_name.upper () == 'ONINIT':
        instance.call (proc, None)

    return instance

  # ---------------------------------------------------------------------------
  # Load data from the database backend
  # ---------------------------------------------------------------------------

  def load (self, classname, obj_id_list, propertylist):

    classdef = self.__getClassdef (classname)

    result = []
    for object_id in obj_id_list:
      if object_id:
        instance = self.__findInstance (classdef, object_id, propertylist)
        result.append (instance.get (propertylist))
      else:
        result.append ([classdef.properties[p].fullType for p in propertylist])

    return result

  # ---------------------------------------------------------------------------
  # Store data in the database backend
  # ---------------------------------------------------------------------------

  def store (self, classname, obj_id_list, propertylist, data):

    opControl = self.__startOperation ('store')

    try:
      classdef = self.__getClassdef (classname)

      result = []
      i = 0
      for object_id in obj_id_list:
        if object_id:
          instance = self.__findInstance (classdef, object_id, [])
          new_object_id = object_id
        else:
          instance = self.__newInstance (classdef)
          new_object_id = instance.get ([u'gnue_id']) [0]
        instance.put (propertylist, data [i])
        i += 1
        result.append (new_object_id)
        self.__dirtyInstances [new_object_id] = instance

    except:
      if opControl:
        self.__cancelOperation ()
      raise

    else:
      if opControl:
        self.__confirmOperation ()

    return result


  # ---------------------------------------------------------------------------
  # Delete business objects
  # ---------------------------------------------------------------------------

  def delete (self, classname, obj_id_list):

    opControl = self.__startOperation ('delete')

    try:
      classdef = self.__getClassdef (classname)

      for objectId in obj_id_list:
        instance = self.__findInstance (classdef, objectId, [])
        for proc in classdef.procedures.values ():
          if proc.gnue_name.upper () == 'ONDELETE':
            instance.call (proc, None)

      for object_id in obj_id_list:
        self.__connection.deleteRecord (classdef.table, object_id)
        if self.__dirtyInstances.has_key (object_id):
          del self.__dirtyInstances [object_id]

    except:
      if opControl:
        self.__cancelOperation ()
      raise

    else:
      if opControl:
        self.__confirmOperation ()


  # ---------------------------------------------------------------------------
  # Call a procedure of business objects
  # ---------------------------------------------------------------------------

  def call (self, classname, obj_id_list, procedurename, parameters):

    opControl = self.__startOperation ('call')

    try:
      classdef = self.__getClassdef (classname)
      procdef  = self.__getProcdef (classdef, procedurename)

      result = []
      for object_id in obj_id_list:
        instance = self.__findInstance (classdef, object_id, [])
        result.append (instance.call (procdef, parameters))

    except:
      if opControl:
        self.__cancelOperation ()

      raise

    else:
      if opControl:
        self.__confirmOperation ()

    return result


  # ---------------------------------------------------------------------------
  # Split a condition into a datasource- and an appserver-condition
  # ---------------------------------------------------------------------------

  def __splitCondition (self, condition, classdef):
    """
    This function splits up a given condition into a condition which is handled
    by the datasource, and another one which has to be processed by appserver.
    """

    if condition is None:
      return (None, None)

    if isinstance (condition, ListType):
      cTree = GConditions.buildTreeFromList (condition)
    else:
      cTree = condition

    dbTrees = []
    asTrees = []
    forest  = []

    if len (cTree._children):
      # NOTE: the first element of a condition tree is a GCondition object
      # self.__splitIntoAnd (cTree._children [0], forest)
      self.__splitIntoAnd (cTree, forest)

    for tree in forest:
      if self.__dbHandlesTree (tree, classdef):
        dbTrees.append (tree)
      else:
        asTrees.append (tree)

    # now create the final trees
    dbCond = None
    asCond = None

    if len (dbTrees):
      dbCond = GConditions.GCondition ()
      comb = GConditions.GCand (dbCond)

      for tree in dbTrees:
        comb._children.append (tree)
        tree._parent = comb

    if len (asTrees):
      asCond = GConditions.GCondition ()
      comb = GConditions.GCand (asCond)

      for tree in asTrees:
        comb._children.append (tree)
        tree._parent = comb

    return (dbCond, asCond)


  # ---------------------------------------------------------------------------
  # Split a condition tree into separate parts
  # ---------------------------------------------------------------------------

  def __splitIntoAnd (self, tree, forest):
    """
    This function splits a condition tree into several independant subtrees
    which can be re-combined via AND.
    """
    if isinstance (tree, GConditions.GConditionElement):
      if isinstance (tree, GConditions.GCand):
        for child in tree._children:
          self.__splitIntoAnd (child, forest)
      else:
        forest.append (tree)
    else:
      for child in tree._children:
        self.__splitIntoAnd (child, forest)


  # ---------------------------------------------------------------------------
  # Determine wether a condition tree can be handled by the datasource or not
  # ---------------------------------------------------------------------------

  def __dbHandlesTree (self, tree, classdef):
    """
    This function traverses the given condition tree and determines if it can
    be handled be the datasource completely.
    """
    if isinstance (tree, GConditions.GCField):
      return self.__usable (tree, classdef)
    else:
      for child in tree._children:
        res = self.__dbHandlesTree (child, classdef)
        if not res:
          return res

    return True


  # ---------------------------------------------------------------------------
  # Return wether a GCField must be handled by appserver or datasource
  # ---------------------------------------------------------------------------
  def __usable (self, element, classdef):
    return not self.__getPropertyDef (element.name, classdef).isCalculated


  # ---------------------------------------------------------------------------
  # Get the property definition for a given property
  # ---------------------------------------------------------------------------

  def __getPropertyDef (self, propertyname, classdef):
    """
    This function gets the property definition for a given propertyname,
    resolving references.
    """
    cDef  = classdef
    parts = propertyname.split ('.')
    for prop in parts [:-1]:
      pDef = cDef.properties [prop]
      if not pDef.isReference:
        raise geasInstance.ReferenceError, (propertyname, prop)
      cDef = pDef.referencedClass

    return cDef.properties [parts [-1]]


  # ---------------------------------------------------------------------------
  # Add all GCField elements of a tree to the content dictionary
  # ---------------------------------------------------------------------------

  def __addConditionToContent (self, classdef, content, tree):
    """
    This function adds all GCField instances in a condition tree to the given
    content dictionary.
    """
    if isinstance (tree, GConditions.GCField):
      self.__getFieldname (classdef, tree.name, content, True)
    else:
      for child in tree._children:
        self.__addConditionToContent (classdef, content, child)

